"""digest and validate configuration files"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from dotenv import load_dotenv
    from importlib.resources import files
    from inspect import getsourcefile
    from os import chmod, path, environ as e, getcwd
    from pathlib import Path
    from posixpath import join as posix_join
    import re
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# TTSPod modules
from logger import Logger
from util import fix_path, check_engines

ENGINES = check_engines()


class Config(object):
    """configuration settings"""
    class Content(object):
        """content processor settings"""

        def __init__(self, working_path=None, log=None):
            self.log = log if log else Logger(debug=True)
            self.attachment_path = path.join(working_path, "attachments")
            self.lua_path = working_path
            self.attachments = e.get('ttspod_attachments')
            if self.attachments and self.attachment_path:
                Path(self.attachment_path).mkdir(parents=True, exist_ok=True)
            return

    class Links(object):
        """link processor settings"""

        def __init__(self, log=None):
            self.log = log if log else Logger(debug=True)
            self.user_agent = e.get('ttspod_user_agent')

    class Wallabag(object):
        """wallabag input settings"""

        def __init__(self, log=None):
            self.log = log if log else Logger(debug=True)
            self.url = e.get('ttspod_wallabag_url')
            self.username = e.get('ttspod_wallabag_username')
            self.password = e.get('ttspod_wallabag_password')
            self.client_id = e.get('ttspod_wallabag_client_id')
            self.client_secret = e.get('ttspod_wallabag_client_secret')

    class Pocket(object):
        """pocket input settings"""

        def __init__(self, log=None):
            self.log = log if log else Logger(debug=True)
            self.consumer_key = e.get('ttspod_pocket_consumer_key')
            self.access_token = e.get('ttspod_pocket_access_token')

    class Insta(object):
        """instapaper input settings"""

        def __init__(self, log=None):
            self.log = log if log else Logger(debug=True)
            self.key = e.get('ttspod_insta_key')
            self.secret = e.get('ttspod_insta_secret')
            self.username = e.get('ttspod_insta_username')
            self.password = e.get('ttspod_insta_password')

    class Pod(object):
        """podcast output settings"""

        def __init__(self, final_path='', ssh_keyfile=None, ssh_password=None, log=None):
            self.log = log if log else Logger(debug=True)
            self.url = posix_join(e.get('ttspod_pod_url'), '')
            self.name = e.get('ttspod_pod_name', 'TTS podcast')
            self.author = e.get('ttspod_pod_author', 'TTS podcast author')
            self.image = e.get('ttspod_pod_image')
            if self.image and not 'http' in self.image:
                self.image = self.url + self.image
            self.description = e.get(
                'ttspod_pod_description', 'TTS podcast description')
            self.language = e.get('ttspod_pod_language', 'en')
            self.ssh_server_path = e.get('ttspod_pod_server_path')
            self.ssh_keyfile = ssh_keyfile
            self.ssh_password = ssh_password
            self.final_path = final_path
            self.rss_file = path.join(final_path, 'index.rss')

    class Speech(object):
        """tts processor settings"""

        def __init__(self, temp_path='./', final_path='./', engine=None,
                     max_workers=10, log=None, debug=False, gpu=1):
            self.log = log if log else Logger(debug=True)
            self.debug = debug
            self.gpu = gpu
            self.engine = engine if engine else e.get('ttspod_engine', '')
            self.engine = self.engine.lower()
            self.eleven_api_key = e.get('ttspod_eleven_api_key')
            self.eleven_voice = e.get('ttspod_eleven_voice', 'Daniel')
            self.eleven_model = e.get(
                'ttspod_eleven_model', 'eleven_monolingual_v1')
            self.openai_api_key = e.get('ttspod_openai_api_key')
            self.openai_voice = e.get('ttspod_openai_voice', 'onyx')
            self.openai_model = e.get('ttspod_openai_model', 'tts-1-hd')
            self.whisper_t2s_model = e.get(
                'ttspod_whisper_t2s_model',
                'whisperspeech/whisperspeech:t2s-fast-medium-en+pl+yt.model')
            self.whisper_s2a_model = e.get(
                'ttspod_whisper_s2a_model',
                'whisperspeech/whisperspeech:s2a-q4-hq-fast-en+pl.model')
            self.voice = e.get('ttspod_voice')
            if self.voice:
                self.voice = path.expanduser(self.voice)
            self.model = e.get(
                'ttspod_model', 'xtts')
            self.model = self.model.lower()
            if not self.voice or self.voice and not path.exists(self.voice):
                if self.engine == 'coqui' and self.model == 'xtts':
                    self.voice = 'Daisy Studious'
                else:
                    self.voice = files('ttspod').joinpath('data', 'sample.wav')
            self.language = e.get('ttspod_language')
            self.max_workers = max_workers
            self.temp_path = fix_path(temp_path, True)
            self.final_path = fix_path(final_path, True)
            if not self.engine:
                self.engine = 'coqui'
            # TODO: some more TTS engine validation
            self.log.write(
                f'Available TTS engines are: {ENGINES}.', log_level=3)
            self.log.write(
                f'TTS settings: engine {self.engine} / model {self.model} / voice {self.voice}',
                log_level=2
            )
            if not self.engine in ENGINES:
                self.log.write(f'TTS engine {self.engine} selected but not available.\n'
                               f'Available engines are: {ENGINES}\n'
                               'reinstall with quickstart.sh to add engines', True)
                self.engine = None

    def __init__(self, debug=True, engine=None, config_path=None, log=None, gpu=None, quiet=False):
        self.log = log if log else Logger(debug=debug)
        self.config_path = None
        if config_path and path.isfile(config_path):
            self.config_path = config_path
        elif (config_path and path.isdir(config_path) and
              path.isfile(path.join(config_path, '.env'))):
            self.config_path = path.join(config_path, '.env')
        elif path.isfile(path.join(Path.home(), '.config', 'ttspod.ini')):
            self.config_path = path.join(Path.home(), '.config', 'ttspod.ini')
        elif path.isfile(path.join(getcwd(), '.config', 'ttspod.ini')):
            self.config_path = path.join(getcwd(), '.config', 'ttspod.ini')
        elif path.isfile(path.join(getcwd(), '.env')):
            self.config_path = path.join(getcwd(), '.env')
        elif path.isfile(path.join(path.dirname(getsourcefile(lambda: 0)), '.env')):
            self.config_path = path.join(
                path.dirname(getsourcefile(lambda: 0)), '.env')
        elif path.isfile(path.join(path.dirname(path.realpath(__file__)), '.env')):
            self.config_path = path.join(
                path.dirname(path.realpath(__file__)), '.env')
        if self.config_path:
            if not quiet:
                self.log.write(
                    f'using stored configuration {self.config_path}', True)
            load_dotenv(self.config_path)
        if not any("ttspod" in x.lower() for x in list(e.keys())):
            raise ValueError(
                'No settings found. Create a .env file or specify the location with --config.'
            )
        if debug is None:
            self.debug = e.get('ttspod_debug', debug)
        else:
            self.debug = debug
        self.log_level = int(e.get('ttspod_log_level', 0))
        self.log.update(debug=self.debug, maximum_level=self.log_level)
        self.gpu = int(gpu if gpu is not None else e.get('ttspod_gpu', 1))
        self.max_length = int(e.get('ttspod_max_length', 20000))
        self.max_workers = int(e.get('ttspod_max_workers', 10))
        self.max_articles = int(e.get('ttspod_max_articles', 5))
        self.working_path = path.join(
            e.get('ttspod_working_path', './working'), '')
        if self.working_path:
            self.working_path = fix_path(self.working_path, True)
        if self.working_path.startswith('./'):
            self.working_path = re.sub(r'^./', '', self.working_path)
            self.working_path = path.join(
                path.dirname(__file__), self.working_path)
        self.temp_path = path.join(self.working_path, 'temp', '')
        self.final_path = path.join(self.working_path, 'output', '')
        self.log_path = e.get('ttspod_log')
        if self.log_path:
            self.log_path = fix_path(self.log_path, False)
        if self.log_path and not '/' in self.log_path and not '\\' in self.log_path:
            self.log_path = path.join(self.working_path, self.log_path)
        if self.log_path:
            self.log.update(debug=self.debug, logfile=self.log_path,
                            maximum_level=self.log_level)
        self.pickle_filename = 'ttspod.pickle'
        self.pickle = path.join(self.working_path, self.pickle_filename)
        if e.get('ttspod_state_file_path'):
            self.state_file_path = posix_join(
                e.get('ttspod_state_file_path'), '')+self.pickle_filename
        else:
            self.state_file_path = None
        if self.state_file_path:
            self.state_file_path = fix_path(self.state_file_path, False)
        self.speech = self.Speech(temp_path=self.temp_path, final_path=self.final_path,
                                  engine=engine, max_workers=self.max_workers, log=self.log,
                                  debug=self.debug, gpu=self.gpu)
        self.content = self.Content(
            working_path=self.working_path, log=self.log)
        self.links = self.Links(log=self.log)
        self.wallabag = self.Wallabag(log=self.log)
        self.pocket = self.Pocket(log=self.log)
        self.insta = self.Insta(log=self.log)
        self.ssh_keyfile = e.get('ttspod_ssh_keyfile')
        self.ssh_password = e.get('ttspod_ssh_password')
        if self.ssh_keyfile:
            self.ssh_keyfile = re.sub(
                r'~/', str(Path.home()).replace('\\', '/') + '/', self.ssh_keyfile)
        if not (self.ssh_keyfile or self.ssh_password):
            key_list = ['id_rsa', 'id_ecdsa', 'id_ecdsa_sk',
                        'id_ed25519', 'id_ed25519_sk', 'id_dsa']
            for key in key_list:
                keyfile = path.join(Path.home(), '.ssh',
                                    key).replace('\\', '/')
                if path.isfile(keyfile):
                    self.ssh_keyfile = keyfile
                    break
        self.pod = self.Pod(
            final_path=self.final_path,
            ssh_keyfile=self.ssh_keyfile,
            ssh_password=self.ssh_password,
            log=self.log
        )
        self.make_files()
        self.validate()

    def validate(self):
        """validate settings and throw error if necessary"""
        if ':' in str(self.state_file_path) or ':' in str(self.pod.ssh_server_path):
            if not self.ssh_keyfile or self.ssh_password:
                raise ValueError(
                    "Remote paths configured for syncing but no SSH keyfile or password provided."
                )
        if self.ssh_keyfile and not path.isfile(self.ssh_keyfile) and not self.ssh_password:
            raise ValueError(
                f"ssh_keyfile {self.ssh_keyfile} does not exist or is not readable."
            )
        if not (
            path.isdir(str(self.working_path)) and
            path.isdir(self.temp_path) and
            path.isdir(self.final_path)
        ):
            raise ValueError(
                f"Unable to access working path {self.working_path}."
            )

    def make_files(self):
        """create working files and directories"""
        Path(self.working_path).mkdir(parents=True, exist_ok=True)
        Path(self.temp_path).mkdir(parents=True, exist_ok=True)
        Path(self.final_path).mkdir(parents=True, exist_ok=True)
        chmod(self.final_path, 0o755)
        if not path.isfile(path.join(self.working_path, 'no_image.lua')):
            with open(path.join(self.working_path, 'no_image.lua'), 'w', encoding='ascii') as f:
                f.write('function Image(el)\nreturn {}\n end')
        return True

    def __str__(self):
        result = (f'config: {str(vars(self))}\nwallabag: {str(vars(self.wallabag))}\n'
                  'pod {str(vars(self.pod))}\nspeech {str(vars(self.speech))}')
        return result



"""python-native highly inefficient implementation of rsync"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

try:
    from getpass import getuser
    from pathlib import Path
    from platform import system
    from posixpath import join as posix_join, split as posix_split
    import hashlib
    import os
    import paramiko
    import re
    import shutil
    import stat
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# TODO: rework the debug/DBG and output logic
# probably want to just generate a string with output and return it
DBG = False


def md5(file_path):
    """Calculate MD5 hash of a file."""
    hash_md5 = hashlib.md5()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

# TODO: this should be refactored to perform the MD5 remotely
# otherwise it's probably not getting any speed benefit


def remote_get_md5(sftp, remote_file):
    """Calculate the MD5 hash of a remote file."""
    try:
        # Open remote file
        with sftp.file(remote_file, "rb") as f:
            hash_md5 = hashlib.md5()
            while True:
                data = f.read(4096)
                if not data:
                    break
                hash_md5.update(data)
            return hash_md5.hexdigest()
    except IOError:
        # If the file doesn't exist on the remote server, return None
        return None


def get_remote_size(sftp, remote_file):
    """Query the file size of a remote file."""
    try:
        info = sftp.stat(remote_file)
        return info.st_size
    except IOError:
        # If the file doesn't exist on the remote server, return None
        return None


def parse_location(fullpath):
    """split path into username, domain name, and filesystem path"""
    if not ':' in fullpath or (system() == 'Windows' and re.match(r'[A-Za-z]:', fullpath)):
        user = None
        domain = None
        path = fullpath
    else:
        m = re.match(
            r'(?P<user>[^@]*(?=@))?@?(?P<domain>[^:]*):(?P<path>.*)?', fullpath)
        user, domain, path = m.group('user'), m.group(
            'domain'), m.group('path')
    return user, domain, path


def remote_isdir(sftp, remote_dir):
    """check if remote path is a directory"""
    if DBG:
        print(f'Checking {remote_dir} for directory status... ', end='')
    try:
        file_attribute = sftp.stat(remote_dir)
        isdir = stat.S_ISDIR(file_attribute.st_mode)
    except Exception:  # pylint: disable=broad-except
        if DBG:
            print(f'Could not check remote dir {remote_dir}.')
        return None
    if isdir:
        if DBG:
            print(f"{remote_dir} is a directory.")
        return True
    else:
        if DBG:
            print(f"{remote_dir} is not a directory.")
        return False


def remote_mkdir(sftp, remote_dir):
    """create remote directory if it does not exist"""
    remote_dir = remote_dir.rstrip('/')
    paths = [remote_dir]
    # sftp mkdir has no -p (parent) mode, so we simulate it by creating each parent directory
    while '/' in remote_dir:
        remote_dir = re.sub(r'/[^/]*$', '', remote_dir)
        paths = [remote_dir] + paths
    try:
        for path in paths:
            if not remote_isdir(sftp, path):
                sftp.mkdir(path)
        return True
    except Exception:  # pylint: disable=broad-except
        if DBG:
            print(f'Could not mkdir {remote_dir}')
        return None


def remote_isfile(sftp, remote_file):
    """check if remote path is a file"""
    try:
        file_attribute = sftp.stat(remote_file)
        isfile = stat.S_ISREG(file_attribute.st_mode)
    except Exception:  # pylint: disable=broad-except
        if DBG:
            print(f'Could not check remote file {remote_file}')
        return None
    if isfile:
        if DBG:
            print(f"{remote_file} is a file")
        return True
    else:
        if DBG:
            print(f"{remote_file} is not a file")
        return False


def remote_get_filelist(sftp=None, remote_dir='', recursive=False):
    """get remote list of files to sync"""
    initial_list = sftp.listdir(path=remote_dir)
    file_list = []
    for file in initial_list:
        if remote_isdir(sftp, posix_join(remote_dir, file)):
            if recursive:
                if DBG:
                    print(f'adding remote {file} directory recursively')
                new_list = remote_get_filelist(sftp=sftp, remote_dir=posix_join(
                    remote_dir, file), recursive=recursive)
                new_list = [posix_join(file, i) for i in new_list]
                file_list.extend(new_list)
            elif DBG:
                print(
                    f'remote {file} is a directory and recursion is disabled, skipping')
        elif remote_isfile(sftp, posix_join(remote_dir, file)):
            file_list.extend([file])
        elif DBG:
            print(f'remote {file} not directory or file, skipping')
    return file_list


def local_get_filelist(local_dir='', recursive=False):
    """build a local list of files to sync"""
    initial_list = os.listdir(local_dir)
    file_list = []
    for file in initial_list:
        if os.path.isdir(os.path.join(local_dir, file)):
            if recursive:
                if DBG:
                    print(f'adding local {file} directory recursively')
                new_list = local_get_filelist(local_dir=os.path.join(
                    local_dir, file), recursive=recursive)
                new_list = [os.path.join(file, i) for i in new_list]
                file_list.extend(new_list)
            elif DBG:
                print(
                    f'local {file} is a directory and recursion is disabled, skipping')
        elif os.path.isfile(os.path.join(local_dir, file)):
            file_list.extend([file])
        elif DBG:
            print(f'local {file} not directory or file, skipping')
    return file_list


def sync(
    source=None,
    destination=None,
    port=22,
    username=None,
    password=None,
    keyfile=None,
    size_only=False,
    debug=False,
    dry_run=False,
    recursive=False
):
    """Sync source folder or file to destination folder."""
    global DBG  # pylint: disable=global-statement
    DBG = debug
    ssh = None
    sftp = None
    source_user, source_host, source_dir = parse_location(source)
    source_dir = source_dir.replace('\\', '/')
    destination_user, destination_host, destination_dir = parse_location(
        destination)
    if not keyfile and not password:
        key_list = ['id_rsa', 'id_ecdsa', 'id_ecdsa_sk',
                    'id_ed25519', 'id_ed25519_sk', 'id_dsa']
        for key in key_list:
            keyfile = os.path.join(Path.home(), '.ssh', key)
            if os.path.isfile(keyfile):
                break
        if not os.path.isfile(keyfile):
            keyfile = None
        elif DBG:
            print(f'Found local ssh keyfile {keyfile}')
    # Establish SSH connection
    if source_host and destination_host:
        raise ValueError("Cannot sync from remote host to remote host")
    if source_host or destination_host:
        host = source_host if source_host else destination_host
        if username:
            user = username
        elif source_host and source_user:
            user = source_user
        elif destination_host and destination_user:
            user = destination_user
        elif getuser():
            user = getuser()
        else:
            raise ValueError(
                "Remote host specified but no username provided or inferred.")
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            if password:
                ssh.connect(host, port, user, password)
            elif keyfile:
                ssh.connect(host, port, user, key_filename=keyfile)
            else:
                raise ValueError(
                    "Either password or ssh key required to make connection")
        except Exception as err:
            raise ValueError(f'SSH connection failed {err}') from err
        if DBG:
            print("connection successful")
        # Open SFTP session
        sftp = ssh.open_sftp()
    else:
        if DBG:
            print("No hostname provided, local sync only.")

    files = []
    source = source.replace('\\', '/')
    source_trail = source.endswith('/')
    destination_trail = destination.endswith('/')
    file_only = False

    if source_host:                                    # remote source, local destination
        # check if source is a file or directory
        if remote_isfile(sftp, source_dir):
            file_only = True
            # extract path and filename
            source_dir, file = posix_split(source_dir)
            # add single file to queue
            files.append(file)
            # if destination ends with /, sync to that directory,
            # otherwise assume destination is filename
            if destination_trail:
                # normalize destination path for POSIX
                destination_dir = os.path.join(destination_dir, '')
        else:                                          # directory transfer
            file_only = False
            # if neither source nor destination ends with /
            # add last part of source dir to destination path
            if not source_trail and not destination_trail:
                destination_dir = os.path.join(
                    destination_dir, os.path.basename(os.path.normpath(source_dir)))
            source_dir = posix_join(source_dir, '')
            files = remote_get_filelist(
                sftp=sftp, remote_dir=source_dir, recursive=recursive)
            destination_dir = os.path.join(destination_dir, '')
        # create destination directory if:
        if not os.path.isdir(destination_dir):
            # there is a leading path in the destination directory
            if file_only and not destination_trail and destination_dir.count('/') > 1:
                parent_dir = re.sub(
                    r'/[^/]*$', '', destination_dir.rstrip('/'))
                try:
                    if not dry_run:
                        Path(parent_dir).mkdir(parents=True, exist_ok=True)
                except Exception as err:
                    raise ValueError(
                        f'Could not create destination directory {parent_dir}: {err}') from err
            # source is directory or source is file and dest ends with /
            if not file_only or (file_only and destination_trail):
                try:
                    if not dry_run:
                        Path(destination_dir).mkdir(
                            parents=True, exist_ok=True)
                except Exception as err:
                    raise ValueError(
                        f'Could not create destination directory {destination_dir}: {err}') from err
    elif destination_host:                               # local source, remote destination
        if os.path.isfile(source_dir):                   # source is file?
            file_only = True
            source_dir, file = os.path.split(
                source_dir)  # extract path and filename
            # add single file to queue
            files.append(file)
            if destination_trail:
                destination_dir = posix_join(destination_dir, '')
        elif os.path.isdir(source_dir):                # source is directory?
            if not source_trail and not destination_trail:
                destination_dir = posix_join(
                    destination_dir, os.path.basename(os.path.normpath(source_dir)))
            source_dir = os.path.join(source_dir, '')
            files = local_get_filelist(
                local_dir=source_dir, recursive=recursive)
            destination_dir = posix_join(destination_dir, '')
        else:                                          # source is neither file nor directory
            raise ValueError(f'Error checking status of source {source_dir}')
        # check if remote directory exists
        if not remote_isdir(sftp, destination_dir):
            if not file_only or (file_only and destination_trail):
                try:
                    # if not, create it
                    remote_mkdir(sftp, destination_dir)
                except Exception as err:
                    raise ValueError(
                        'Could not create remote destination directory '
                        f'{destination_dir}: {err}') from err
    else:                                                  # source and destination both local
        # source is file, not directory
        if os.path.isfile(source_dir):
            file_only = True
            source_dir, file = os.path.split(
                source_dir)   # extract path and filename
            # add single file to queue
            files.append(file)
            if destination_trail:  # if destination ends with /, assume it is a dir
                destination_dir = os.path.join(destination_dir, '')
        else:  # source is directory
            if not source_trail and not destination_trail:
                destination_dir = os.path.join(
                    destination_dir, os.path.basename(os.path.normpath(source_dir)))
            source_dir = os.path.join(source_dir, '')
            files = local_get_filelist(source_dir, recursive=recursive)
            destination_dir = os.path.join(destination_dir, '')
        if not os.path.isdir(destination_dir):
            if not file_only or (file_only and destination_trail):
                try:
                    if not dry_run:
                        Path(destination_dir).mkdir(
                            parents=True, exist_ok=True)
                except Exception as err:
                    raise ValueError(
                        'Could not create destination directory '
                        f'{destination_dir}: {err}') from err

    if DBG:
        print(
            f'source host: {source_host}\n'
            f'source: {source_dir}\n'
            f'destination host: {destination_host}\n'
            f'destination: {destination_dir}'
        )

    # Sync files
    if destination_host:  # local source to remote destination
        for filename in files:
            # local system can be Linux/Mac/Windows
            local_file = os.path.join(source_dir, filename)
            if filename.count('/'):
                local_file_path = re.sub(r'/[^/]*$', '', filename)
                if not dry_run:
                    remote_mkdir(sftp, posix_join(
                        destination_dir, local_file_path))
            if file_only and not destination_trail:
                remote_file = destination_dir
            else:
                # assume remote system supports POSIX paths
                remote_file = posix_join(destination_dir, filename)
            if os.path.isfile(local_file):
                local_check = os.path.getsize(
                    local_file) if size_only else md5(local_file)
                if DBG:
                    print(f'local {local_check}... ', end='')
                remote_check = get_remote_size(
                    sftp, remote_file) if size_only else remote_get_md5(sftp, remote_file)
                if DBG:
                    print(f'remote {remote_check}... ', end='')
                if remote_check is None:
                    if DBG:
                        print(
                            f"File {filename} does not exist on the remote server. Uploading...")
                    if not dry_run:
                        local_stat = os.stat(local_file)
                        times = (local_stat.st_atime, local_stat.st_mtime)
                        sftp.put(local_file, remote_file)
                        sftp.utime(remote_file, times)
                        sftp.chmod(remote_file, local_stat.st_mode)
                elif local_check != remote_check:
                    if DBG:
                        print(
                            f"File {filename} is different. Uploading updated version...")
                    if not dry_run:
                        local_stat = os.stat(local_file)
                        times = (local_stat.st_atime, local_stat.st_mtime)
                        sftp.put(local_file, remote_file)
                        sftp.utime(remote_file, times)
                        sftp.chmod(remote_file, local_stat.st_mode)
                else:
                    if DBG:
                        print(
                            f"File {filename} is identical. No need to upload.")
            elif DBG:
                print(f"Could not find source file {filename}.")
    elif source_host:  # remote source to local destination
        for filename in files:
            # assume remote system supports POSIX paths
            remote_file = posix_join(source_dir, filename)
            if not file_only and remote_file.count('/') > 1:
                remote_file_path = re.sub(r'^[^/]*/', '', remote_file)
                remote_file_path = re.sub(r'/[^/]*$', '', remote_file_path)
                if not dry_run:
                    Path(os.path.join(destination_dir, remote_file_path)).mkdir(
                        parents=True, exist_ok=True)
            local_file = os.path.join(
                destination_dir, filename
            ) if not file_only or destination_trail else destination_dir
            local_check = None
            if os.path.isfile(local_file):
                remote_check = get_remote_size(
                    sftp, remote_file) if size_only else remote_get_md5(sftp, remote_file)
                if DBG:
                    print(f'remote {remote_check}... ', end='')
                local_check = os.path.getsize(
                    local_file) if size_only else md5(local_file)
                if DBG:
                    print(f'local {local_check}... ', end='')
                if local_check != remote_check:
                    if DBG:
                        print(
                            f"File {filename} is different. Downloading updated version...")
                    if not dry_run:
                        remote_stat = sftp.stat(remote_file)
                        times = (remote_stat.st_atime, remote_stat.st_mtime)
                        sftp.get(remote_file, local_file)
                        os.utime(local_file, times)
                else:
                    if DBG:
                        print(
                            f"File {filename} is identical. No need to download.")
            else:
                if DBG:
                    print(
                        f"File {filename} does not exist on the local server. Downloading...")
                if not dry_run:
                    remote_stat = sftp.stat(remote_file)
                    times = (remote_stat.st_atime, remote_stat.st_mtime)
                    sftp.get(remote_file, local_file)
                    os.utime(local_file, times)
    else:  # local source to local destination
        for filename in files:
            local_file = os.path.join(source_dir, filename)
            remote_file = os.path.join(destination_dir, filename)
            local_check = None
            if os.path.isfile(local_file) and os.path.isfile(remote_file):
                local_check = os.path.getsize(
                    local_file) if size_only else md5(local_file)
                if DBG:
                    print(f'local {local_check}')
                remote_check = os.path.getsize(
                    remote_file) if size_only else md5(remote_file)
                if DBG:
                    print(f'remote {remote_check}')
                if local_check != remote_check:
                    if DBG:
                        print(
                            f"File {filename} is different. Copying updated version...")
                    try:
                        if not dry_run:
                            shutil.copy2(local_file, remote_file)
                    except Exception as err:  # pylint: disable=broad-except
                        print(f'Copy failed with error {err}')
                else:
                    if DBG:
                        print(
                            f"File {filename} is identical. No need to copy.")
            else:
                if DBG:
                    print(
                        f"File {filename} does not exist on the destination. Copying...")
                try:
                    if not dry_run:
                        remote_file_path = os.path.dirname(remote_file)
                        if not os.path.isdir(remote_file_path):
                            Path(remote_file_path).mkdir(
                                parents=True, exist_ok=True)
                        shutil.copy2(local_file, remote_file)
                except Exception as err:  # pylint: disable=broad-except
                    print(f'Copy failed with error {err}')

    if sftp and ssh:
        # Close connection
        sftp.close()
        ssh.close()



"""freestanding script to generate sample audio output"""
# optional system certificate trust
# cspell: disable
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from os import path, getcwd, listdir
    from pathlib import Path
    from shutil import rmtree
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# TTSPod modules
from speech import speech

TEXT = """
    The quick brown fox jumped over the lazy dog. The dog was so lazy, he did not even notice that the fox was brown.
    Have you ever seen a blue fox? I have not!
    """


class Config(object):
    """simplified config option for generating speech samples"""

    def __init__(self):
        self.engine = "coqui"
        self.nltk = True
        self.log = None
        self.gpu = 1
        self.tortoise_path = path.join(
            getcwd(), 'examples/voices/tortoise/', '')
        self.xtts_path = path.join(getcwd(), 'examples/voices/xtts', '')
        self.final_path = None
        self.model = None
        self.voice = None


def main():
    """generate sample TTS files"""

    c = Config()
    for p in [c.tortoise_path, c.xtts_path]:
        if path.exists(p):
            print(f'Output folder {p} already exists.\n')
            remove_question = input('Remove and recreate? (y/n)')
            if remove_question.lower() == "y":
                rmtree(p)

    c.final_path = c.tortoise_path
    c.model = "tortoise"
    Path(c.final_path).mkdir(parents=True, exist_ok=True)

    voice_path = path.join(getcwd(), 'working/voices', '')
    if not path.isdir(voice_path):
        print(f'{voice_path} does not exist.\n')
        path_question = input('Enter path for voice samples to generate:')
        voice_path = path.join(path.expanduser(path_question), '')
        if not path.isdir(voice_path):
            print(f'Cannot find path {voice_path}. Exiting.')
            exit()

    voices = listdir(voice_path)
    voices.sort()

    print(f'sample text:\n{TEXT}')
    for voice in voices:
        try:
            print(f'generating sample {voice}')
            c.voice = path.join(voice_path, voice, '')
            tts = speech.Speech(config=c)
            fullpath = tts.speechify(
                title=voice, raw_text=TEXT, overwrite=False)
            print(f'output to: {fullpath}')
            del tts
        except Exception as err:  # pylint: disable=broad-except
            print(f'failed with {err}\n')

    c.final_path = c.xtts_path
    c.model = "xtts"
    Path(c.final_path).mkdir(parents=True, exist_ok=True)

    voices = [
        'Claribel Dervla', 'Daisy Studious', 'Gracie Wise', 'Tammie Ema', 'Alison Dietlinde',
        'Ana Florence', 'Annmarie Nele', 'Asya Anara', 'Brenda Stern', 'Gitta Nikolina',
        'Henriette Usha', 'Sofia Hellen', 'Tammy Grit', 'Tanja Adelina', 'Vjollca Johnnie',
        'Andrew Chipper', 'Badr Odhiambo', 'Dionisio Schuyler', 'Royston Min', 'Viktor Eka',
        'Abrahan Mack', 'Adde Michal', 'Baldur Sanjin', 'Craig Gutsy', 'Damien Black',
        'Gilberto Mathias', 'Ilkin Urbano', 'Kazuhiko Atallah', 'Ludvig Milivoj', 'Suad Qasim',
        'Torcull Diarmuid', 'Viktor Menelaos', 'Zacharie Aimilios', 'Nova Hogarth', 'Maja Ruoho',
        'Uta Obando', 'Lidiya Szekeres', 'Chandra MacFarland', 'Szofi Granger',
        'Camilla Holmström', 'Lilya Stainthorpe', 'Zofija Kendrick', 'Narelle Moon',
        'Barbora MacLean', 'Alexandra Hisakawa', 'Alma María', 'Rosemary Okafor',
        'Ige Behringer', 'Filip Traverse', 'Damjan Chapman', 'Wulf Carlevaro', 'Aaron Dreschner',
        'Kumar Dahl', 'Eugenio Mataracı', 'Ferran Simen', 'Xavier Hayasaka', 'Luis Moray',
        'Marcos Rudaski'
    ]

    for voice in voices:
        try:
            print(f'generating sample {voice}')
            c.voice = voice
            tts = speech.Speech(config=c)
            fullpath = tts.speechify(voice, TEXT)
            del tts
        except Exception as err:  # pylint: disable=broad-except
            print(f'failed with {err}\n')


if __name__ == "__main__":
    main()



"""main TTS processor"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass

# standard modules
try:
    from warnings import simplefilter
    import os
    import re
    from time import time
    import unicodedata
    import uuid
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# TTSPod modules
from logger import Logger
from util import check_engines

simplefilter(action='ignore', category=FutureWarning)

ENGINES = check_engines()


class Speech(object):
    """main TTS processor"""

    def __init__(self, config, dry=False, log=None):
        self.dry = dry
        if dry:
            return
        self.log = log if log else Logger(debug=True)
        self.config = config
        self.config.nltk = False
        self.final_path = config.final_path
        # pylint: disable=import-outside-toplevel
        match self.config.engine.lower():
            case "openai" if "openai" in ENGINES:
                from paid import Paid
                self.tts = Paid(config=self.config, log=self.log)
            case "eleven" if "eleven" in ENGINES:
                from paid import Paid
                self.tts = Paid(config=self.config, log=self.log)
            case "whisper" if "whisper" in ENGINES:
                from whisper import Whisper
                self.tts = Whisper(config=self.config, log=self.log)
            case "coqui" if "coqui" in ENGINES:
                from coqui import Coqui
                self.tts = Coqui(config=self.config, log=self.log)
            case "f5" if "f5" in ENGINES:
                from f5 import F5
                self.tts = F5(config=self.config, log=self.log)
            case _:
                raise ValueError('TTS engine not configured')
        # pylint: enable=import-outside-toplevel
    def slugify(self, value):
        """convert an arbitrary string to a valid filename"""
        value = str(value)
        value = unicodedata.normalize('NFKD', value).encode(
            'ascii', 'ignore').decode('ascii')
        value = re.sub(r'[^\w\s-]', '', value.lower())
        return re.sub(r'[-\s]+', '-', value).strip('-_')

    def speechify(self, title="No Title Available", text="", overwrite=True):
        """workhorse TTS function"""
        clean_title = self.slugify(title)
        out_file = os.path.join(self.config.final_path, f'{clean_title}.mp3')
        if os.path.isfile(out_file) and not overwrite:
            self.log.write(f'skipping {out_file} because it already exists')
            return out_file

        temp = str(uuid.uuid4())
        start_time = time()
        if os.path.exists(out_file):  # don't overwrite existing files
            out_file = os.path.join(
                self.config.final_path, f'{clean_title}-{temp}.mp3')

        if self.dry:  # quit if dry run
            self.log.write(f'dry run: not creating {out_file}')
            return
        self.log.write(f'starting TTS conversion to {out_file}')
        if title != "No Title Available":
            text = title.strip() + ".\n\n" + text.strip()
        self.log.write(self.tts.convert(text=text, output_file=out_file))
        elapsed = round(time() - start_time)
        self.log.write(
            f'TTS conversion of {out_file} complete, elapsed time: {elapsed} seconds')

        if os.path.isfile(out_file):
            os.chmod(out_file, 0o644)
            return out_file
        else:
            self.log.write(f'TTS conversion of {out_file} failed.')
            return None



"""generate audio using whisperspeech model"""
# optional system certificate trust
try:
    import truststore
    truststore.inject_into_ssl()
except ImportError:
    pass
try:
    from warnings import simplefilter
    from contextlib import redirect_stdout, redirect_stderr
    from glob import glob
    from os import path, environ as env
    from pprint import pprint
    import torch
    import torchaudio
    from pathlib import Path
    from platform import processor
    from transformers import pytorch_utils
    from whisperspeech.pipeline import Pipeline
    from io import StringIO
except ImportError as e:
    print(
        f'Failed to import required module: {e}\n'
        'You may need to re-execute quickstart.sh.\n'
        'See https://github.com/ajkessel/ttspod/blob/main/README.md for details.')
    exit()

# ttspod modules
from logger import Logger
from util import patched_isin_mps_friendly, chunk

# suppress spurious UserWarning from Whisper
simplefilter(action='ignore', category=UserWarning)
simplefilter(action='ignore', category=FutureWarning)

# this attempts to minimize random voice variations
torch.manual_seed(123456789)

# sensible default settings if none are provided
DEVICE = 'cpu'
TEMPERATURE = 0.2

if torch.cuda.is_available():
    DEVICE = "cuda"
elif torch.backends.mps.is_available() and processor() != 'i386':
    DEVICE = 'mps'
    env["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"
    pytorch_utils.isin_mps_friendly = patched_isin_mps_friendly

# cspell: disable
if "cuda" in DEVICE and torch.cuda.get_device_name().endswith("[ZLUDA]"):
    torch.backends.cudnn.enabled = False
    torch.backends.cuda.enable_flash_sdp(False)
    torch.backends.cuda.enable_math_sdp(True)
    torch.backends.cuda.enable_mem_efficient_sdp(False)
# cspell: enable


class Whisper:
    """whisper text to speech generator"""

    def __init__(
        self, config=None, log=None,
        t2s_model=None, s2a_model=None, voice=None, gpu='gpu'
    ) -> None:
        self.log = log if log else Logger(debug=True)
        self.config = config
        self.voice = None
        if not config:
            c = {}
        else:
            if not isinstance(config, dict):
                c = vars(config)
            else:
                c = config
        if gpu == 'cpu':
            self.gpu = 'cpu'
        else:
            self.gpu = DEVICE
        t2s_model = c.get(
            'whisper_t2s_model', 'whisperspeech/whisperspeech:t2s-base-en+pl.model')
        s2a_model = c.get(
            'whisper_a2s_model', 'whisperspeech/whisperspeech:s2a-q4-base-en+pl.model')
        voice = path.expanduser(voice if voice else c.get('voice', ''))
        if path.isfile(voice):
            self.voice = voice
        elif path.isdir(voice):
            audio_files = glob(path.join(voice, "*wav")) + \
                glob(path.join(voice, "*mp3"))
            if audio_files:
                self.voice = audio_files[0]
        self.tts = Pipeline(t2s_ref=t2s_model,
                            s2a_ref=s2a_model,
                            device=self.gpu,
                            torch_compile=True,
                            optimize=True)
        self.log.write('Whisper generator initialized.',
                       error=False, log_level=2)

    def generate(self, texts=None, cps=15, output=None, speaker=None):
        """main whisperspeech generator"""
        if not speaker:
            self.log.write('using default speaker')
            speaker = self.tts.default_speaker
        elif isinstance(speaker, (str, Path)):
            self.log.write(f'extracting speaker {speaker}')
            speaker = self.tts.extract_spk_emb(speaker)
        generated = []
        stdout_buffer = StringIO()
        stderr_buffer = StringIO()
        stoks = torch.empty((0, 0), dtype=float)
        atoks = stoks
        old_stoks = stoks
        old_atoks = stoks
        atoks_prompt = None
        text = ""
        for i, next_text in enumerate(texts):
            self.log.write(
                f'Processing chunk {i+1} of {len(texts)}:\n{next_text}',
                error=False,
                log_level=3)
            if i == len(texts)-1:
                text += f' {next_text}'
            elif len(text) + len(next_text) < 300:
                text += f' {next_text}'
                continue
            if not text:
                continue
            self.log.write(
                f'Chunked together as:\n{text}', error=False, log_level=3)
            try:
                with redirect_stdout(stdout_buffer), redirect_stderr(stderr_buffer):
                    stoks = self.tts.t2s.generate(
                        txt=text,
                        cps=cps,
                        lang='en',
                        T=TEMPERATURE,
                        stoks_prompt=None,
                        show_progress_bar=False
                    )[0]
                    stoks = stoks[stoks != 512]
                    if len(old_stoks) > 0 and len(old_atoks) > 0:
                        stoks = torch.cat([old_stoks[-100:], stoks])
                        atoks_prompt = old_atoks[:, :, -300:]
                    atoks = self.tts.s2a.generate(
                        stoks=stoks,
                        speakers=speaker.unsqueeze(0),
                        atoks_prompt=atoks_prompt,
                        show_progress_bar=False,
                        T=TEMPERATURE
                    )
                if atoks_prompt is not None:
                    atoks = atoks[:, :, 301:]
                old_stoks = stoks
                old_atoks = atoks
                generated.append(atoks)
            except Exception as err:  # pylint: disable=broad-except
                self.log.write(f'Something went wrong: {err}')
                old_stoks = torch.empty((0, 0), dtype=float)
                old_atoks = old_stoks
            text = next_text
        result = stdout_buffer.getvalue()+"\n"+stderr_buffer.getvalue()
        try:
            audios = []
            for i, atok in enumerate(generated):
                if i != 0:
                    audios.append(torch.zeros((1, int(24000*0.5)),
                                              dtype=atok.dtype, device=self.gpu))
                audios.append(self.tts.vocoder.decode(atok))
            if output:
                torchaudio.save(output, torch.cat(audios, -1).cpu(), 24000)
                if path.isfile(output):
                    return True
        except Exception as err:  # pylint: disable=broad-except
            self.log.write(f'Something went wrong: {err}\n{result}')

    def convert(self, text, output_file):
        """convert text input to given output_file"""
        chunks = chunk(text)
        try:
            results = self.generate(
                texts=chunks, output=output_file, speaker=self.voice)
            if results:
                self.log.write(
                    f'TTS conversion complete: {results}')
                return True
            else:
                return False
        except Exception as err:  # pylint: disable=broad-except
            self.log.write(f'TTS conversion failed: {err}', True)
            return False


if __name__ == "__main__":
    tts = Whisper()
    pprint(vars(tts))
    pprint(dir(tts))

    # pylint: disable=line-too-long
#     TEXT = """A Hare was making fun of the Tortoise one day for being so slow.
#     "Do you ever get anywhere?" he asked with a mocking laugh.
#     "Yes," replied the Tortoise, "and I get there sooner than you think. I'll run you a race and prove it."
#     The Hare was much amused at the idea of running a race with the Tortoise, but for the fun of the thing he agreed.
#     So the Fox, who had consented to act as judge, marked the distance and started the runners off.
#     The Hare was soon far out of sight, and to make the Tortoise feel very deeply how ridiculous it was for him to try a race with a Hare, he lay down beside the course to take a nap until the Tortoise should catch up.
#     The Tortoise meanwhile kept going slowly but steadily, and, after a time, passed the place where the Hare was sleeping. But the Hare slept on very peacefully; and when at last he did wake up, the Tortoise was near the goal. The Hare now ran his swiftest, but he could not overtake the Tortoise in time.
# """
#     tts = Whisper(voice='~/ttspod/working/voices/it')
#     tts.convert(TEXT, "whisper-test.mp3")
